using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Drawing.Text;
using System.Text;
using System.Windows.Forms;
using System.Diagnostics;

namespace TextFormatFlagsUtil
{
    public partial class OptionsForm : Form
    {
        /// <summary>
        ///     Cached shared arrays containing the name of the different StringFormat options.
        ///     Used when setting/getting the values in/from the StringFormat option boxes.
        /// </summary>
        static readonly string[] textFlagNames;
        static readonly string[] stringFlagNames;
        static readonly string[] alignmentNames;
        static readonly string[] lineAlignNames;
        static readonly string[] trimNames;
        static readonly string[] prefixNames;

        /// <summary>
        ///     The current options and objects in the OptionsForm.
        /// </summary>
        TextFormatFlags flags;
        StringFormat format;
        Font drawingFont;
        string drawingText;
        MsdnForm msdn;
        SurfaceForm drawingSurface;
        SizeConverter sizeConverter = new SizeConverter();

        /// <summary>
        ///     Flag used when updating options boxes to prevent the OptionsChanged event from raising.
        /// </summary>
        bool suspendEvents;

        /// <summary>
        ///     Delegate and event definition for the OptionsChanged event.    
        /// </summary>
        /// <param name="optionType"></param>
        /// <param name="newValue"></param>
        public delegate void OptionsChangedCallback(OptionType optionType, object newValue);
        private event OptionsChangedCallback optionsChangedHandlers;

        /// <summary>
        ///     Class created for showing MSDN documentation.
        /// </summary>
        internal class MsdnForm : Form
        {
            public WebBrowser browser;

            public MsdnForm()
            { 
                browser = new WebBrowser();
                browser.Dock = DockStyle.Fill;
                this.Controls.Add(browser);
                this.Size = new Size(800, 600);
            }
        }

        /// <summary>
        ///     Enum defining the different options that can be changed from the Options form.
        /// </summary>
        public enum OptionType
        { 
            Font,
            Text,
            TextFormatFlags,
            StringFormat,
            AutoSize,
            UseBindingBox
        }

        /// <summary>
        ///     OptionsForm class constructor.
        /// </summary>
        static OptionsForm()
        {
            textFlagNames = Enum.GetNames(typeof(TextFormatFlags));
            stringFlagNames = Enum.GetNames(typeof(StringFormatFlags));
            alignmentNames = Enum.GetNames(typeof(StringAlignment));
            lineAlignNames = Enum.GetNames(typeof(StringAlignment));
            trimNames = Enum.GetNames(typeof(StringTrimming));
            prefixNames = Enum.GetNames(typeof(HotkeyPrefix));
        }

        /// <summary>
        ///     OptionsForm instance constructor.
        /// </summary>
        public OptionsForm()
        {
            InitializeComponent();

            HookupBoxes(this.groupBoxTextFormatFlags);
            HookupBoxes(this.groupBoxStringFormat);

            this.flags = TextFormatFlags.Default;
            this.format = StringFormat.GenericDefault;

            UpdateStringFormatBoxes();
            UpdateTextFormatFlagsBoxes();

            ClearAutoSizeBoxInfo();

            this.drawingFont = this.Font;
            // DrawingSurface initialized on demand.
        }

        /// <summary>
        ///     This method installs the CheckChanged event in the different options boxes (CheckBox/RadioButton objects).
        /// </summary>
        /// <param name="container"></param>
        private void HookupBoxes(GroupBox container)
        {
            if (container == null)
            {
                return;
            }

            foreach (Control c in container.Controls)
            {
                if (c is GroupBox)
                {
                    HookupBoxes(c as GroupBox);
                }

                CheckBox cb = c as CheckBox;

                if (cb != null)
                {
                    cb.CheckedChanged += new EventHandler(Box_CheckedChanged);
                }
                else
                {
                    RadioButton rb = c as RadioButton;

                    if (rb != null)
                    {
                        rb.CheckedChanged += new EventHandler(Box_CheckedChanged);
                    }
                }

            }
        }

        
        /// <summary>
        ///     Raises the OptionsChanged event.
        /// </summary>
        /// <param name="optionType">The type of the option that changed</param>
        /// <param name="newValue">The new value of the changing option</param>
        private void OnOptionsChanged(OptionType optionType, object newValue)
        {
            if (optionsChangedHandlers != null)
            {
                optionsChangedHandlers(optionType, newValue);
            }
        }

        // Properties

        /// <summary>
        ///     Current selected font.
        /// </summary>
        private Font DrawingFont
        {
            get
            {
                return this.drawingFont;
            }
            set
            {
                if (value != null && !value.Equals(this.drawingFont))
                {
                    this.drawingFont = value;
                    OnDrawingFontChanged();
                }
            }
        
        }

        /// <summary>
        ///     Drawing surface Window.
        /// </summary>
        private SurfaceForm DrawingSurface
        {
            get
            {
                if (this.drawingSurface == null || this.drawingSurface.IsDisposed)
                {
                    InitializeDrawingForm();
                }

                return this.drawingSurface;
            }
        }
        

        /// <summary>
        ///     OptionsChanged event.
        /// </summary>
        public event OptionsChangedCallback OptionsChanged
        {
            add
            {
                optionsChangedHandlers += value;
            }
            remove 
            {
                optionsChangedHandlers -= value;
            }
        }

        ///// Events

        /// <summary>
        ///     Event raised when the selected font is changed in the OptionsForm.
        /// </summary>
        private void OnDrawingFontChanged()
        {
            this.linkLabelFont.Text = String.Format("&Font: Name={0}, Style={1}, Size={2}", this.drawingFont.Name, this.drawingFont.Style, this.drawingFont.SizeInPoints);
            Font txtBoxFont = this.textBoxDrawingText.Font;
            this.textBoxDrawingText.Font = new Font(this.drawingFont.FontFamily, txtBoxFont.Size, txtBoxFont.Style, txtBoxFont.Unit);
            OnOptionsChanged(OptionType.Font, this.drawingFont);
        }

        /// <summary>
        ///     Event raised when the AutoSize option is changed in the OptionsForm.
        /// </summary>
        private void OnDrawingSurfaceAutoSizeChanged()
        {
            OnOptionsChanged(OptionType.AutoSize, this.checkBoxAutoSize.Checked);

            if (!this.checkBoxAutoSize.Checked)
            {
                ClearAutoSizeBoxInfo();
            }
        }

        /// <summary>
        ///     Event raised when the text to be drawn is changed in the OptionsForm.
        /// </summary>
        private void OnDrawingTextChanged()
        {
            OnOptionsChanged(OptionType.Text, this.drawingText);
        }

        /// <summary>
        ///     Event raised when one of the StringFormat options in changed in the OptionsForm.
        ///     This method updates the current StringFormat object and triggers the OptionsChanged event.
        /// </summary>
        /// <param name="sender"></param>
        private void OnStringFormatChanged(object sender)
        {
            // Determine what option changed in the StringFormat options and parse its value and set the 
            // StringFormat object accordingly.

            CheckBox cb = sender as CheckBox;

            if (cb != null)
            {
                foreach (string flagName in stringFlagNames)
                {
                    if (cb.Text == flagName) // Option found.
                    {

                        StringFormatFlags flag = (StringFormatFlags)Enum.Parse(typeof(StringFormatFlags), flagName);
                        
                        if (cb.Checked)
                        {
                            this.format.FormatFlags |= flag;
                        }
                        else
                        {
                            this.format.FormatFlags &= ~flag;
                        }
                        
                        OnOptionsChanged(OptionType.StringFormat, this.format);
                        break;
                    }
                }
            }
            else
            {
                RadioButton rb = sender as RadioButton;

                if( rb == null )
                {
                    Debug.Fail("Sender object not a CheckBox or RadioButton");
                    return;
                }

                switch( rb.Parent.Text ) // GroupBox.Text
                {
                    case "LineAlignment":
                        this.format.LineAlignment = (StringAlignment)Enum.Parse(typeof(StringAlignment), rb.Text);
                        break;
                    case "Alignment":
                        this.format.Alignment = (StringAlignment)Enum.Parse(typeof(StringAlignment), rb.Text);
                        break;
                    case "HotkeyPrefix":
                        this.format.HotkeyPrefix = (HotkeyPrefix)Enum.Parse(typeof(HotkeyPrefix), rb.Text);
                        break;
                    case "Trimming":
                        this.format.Trimming = (StringTrimming)Enum.Parse(typeof(StringTrimming), rb.Text);
                        break;
                    default:
                        Debug.Fail("rb.Parent.Text unexpected: " + rb.Parent.Text);
                        break;
                }
            }

            OnOptionsChanged(OptionType.StringFormat, this.format);
        }

        /// <summary>
        ///     Event raised when the StringFormat option boxes in the OptionsForm need to be updated according 
        ///     to the specified StringFormat object.
        /// </summary>
        private void OnStringFormatUpdate(StringFormat newFormat)
        {
            Debug.Assert(newFormat != null, "newFormat should not be null.");

            if (!this.format.Equals(newFormat))
            {
                this.format.Dispose();
                this.format = newFormat;
                UpdateStringFormatBoxes();

                OnOptionsChanged(OptionType.StringFormat, this.format);
            }
        }

        /// <summary>
        ///     Event raised when one of the TextFormatFlags flags is changed in the OptionsForm.
        ///     This method updates the current TextForamtFlags flags and triggers the OptionsChanged event.
        /// </summary>
        /// <param name="sender"></param>
        private void OnTextFormatFlagsChanged(object sender)
        {
            // Determine what option changed in the TextFormatFlags options and parse its value and set the 
            // TextFormatFlags object accordingly.

            CheckBox cb = sender as CheckBox;

            if (cb != null)
            {
                foreach (string flagName in textFlagNames)
                {
                    if (cb.Text == flagName)
                    {
                        TextFormatFlags flag = (TextFormatFlags)Enum.Parse(typeof(TextFormatFlags), flagName);

                        if (cb.Checked)
                        {
                            this.flags |= flag;
                        }
                        else
                        {
                            this.flags &= ~flag;
                        }

                        OnOptionsChanged(OptionType.TextFormatFlags, this.flags);

                        break;
                    }
                }
            }
        }

        /// <summary>
        ///     Event raised when the TextFormatFlags option boxes in the OptionsForm need to be updated according 
        ///     to the value of the TextFormatFlags object.
        /// </summary>
        private void OnTextFormatFlagsUpdate(TextFormatFlags newFlags)
        {
            if (this.flags != newFlags)
            {
                this.flags = newFlags;
                UpdateTextFormatFlagsBoxes();

                OnOptionsChanged(OptionType.TextFormatFlags, this.flags);
            }
        }

        /// <summary>
        ///     Event raised when the 'DrawSurface as binding box' option is changed.
        /// </summary>
        private void OnUseBindingBoxChanged()
        {
            OnOptionsChanged(OptionType.UseBindingBox, this.checkBoxUseBindingBox.Checked);
        }


        // Controls event handlers

        /// <summary>
        ///     Shows the drawing surface window if hidden.
        /// </summary>
        private void ButtonShowSurface_Click(object sender, EventArgs e)
        {
            this.DrawingSurface.Show();
        }

        /// <summary>
        ///     Forces an update of (Synchronizes) the StringFormat options based on the current TextFormatFlags options.
        /// </summary>
        private void ButtonUpdateStringFormat_Click(object sender, EventArgs e)
        {
            StringFormat newFormat = GetStringFormatFromTextFormatFlags(this.flags);
            OnStringFormatUpdate(newFormat);
        }

        /// <summary>
        ///     Forces an update of (Synchronizes) the TextFormatFlags options based on the current StringFormat options.
        /// </summary>
        private void ButtonUpdateTextFormat_Click(object sender, EventArgs e)
        {
            TextFormatFlags newFlags = GetTextFormatFlagsFromStringFormat(this.format);
            OnTextFormatFlagsUpdate(newFlags);
        }

        /// <summary>
        ///     Raises the DrawingTextChanged event to force an update of the current text in the drawing surface.
        /// </summary>
        private void ButtonUpdateText_Click(object sender, EventArgs e)
        {
            this.drawingText = this.textBoxDrawingText.Text;
            OnDrawingTextChanged();
        }

        /// <summary>
        ///     Event handler for the different option boxes in the OptionsForm form.  This method raises the 
        ///     appropriate event according to the option that has changed.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void Box_CheckedChanged(object sender, EventArgs e)
        {
            if (suspendEvents)
            {
                return;
            }

            Control container = sender as Control;

            for(;;)
            {
                container = container.Parent;

                if (container == null)
                {
                    break;
                }

                if (container.Text == "TextFormatFlags")
                {
                    OnTextFormatFlagsChanged(sender);
                    break;
                }
                else if (container.Text == "StringFormat")
                {
                    OnStringFormatChanged(sender);
                    break;
                }
            }
        }

        /// <summary>
        ///     Raises the 'Draw AutoSize Rectangle' event - this forces the drawing surface
        ///     to calculate and draw the preferred bounding box of the text.
        /// </summary>
        private void CheckBoxAutoSize_CheckedChanged(object sender, EventArgs e)
        {
            OnDrawingSurfaceAutoSizeChanged();
        }

        /// <summary>
        ///     Raise the 'Use Drawing Surface as Bounding Box' event - this forces the drawing surface
        ///     to use the Panel canvas area as the text bounding box, changing the size of the surface
        ///     will affect this bounding box and probably the text layout.        
        /// </summary>
        private void CheckBoxUseBindingBox_CheckedChanged(object sender, EventArgs e)
        {
            OnUseBindingBoxChanged();
        }

        /// <summary>
        ///     Event handler for the DrawingSurface.AutoSizeBoxChanged event.  This event is raised
        ///     when the text auto-size bounding box has been painted.  The handler updates the auto-size
        ///     boxes in the OptionsForm form with the new information.
        /// </summary>
        void DrawingSurface_AutoSizeBoxChanged(SurfaceForm.SurfaceType surfaceType, Size newSize)
        {
            string newSizeStr = newSize.Width + ", " + newSize.Height;
            switch (surfaceType)
            {
                case SurfaceForm.SurfaceType.Graphics:
                    this.txtBoxDrawStringInfo.Text = newSizeStr;
                    break;
                case SurfaceForm.SurfaceType.TextRenderer:
                    this.txtBoxDrawTextInfo.Text = newSizeStr;
                    break;
            }
        }

        /// <summary>
        ///     Set the UStringFormat object to StringFormat.GenericDefault and forces an update of the
        ///     StringFormat option boxes in the OptionsForm.
        /// </summary>
        private void BtnStringFormatGenericDefault_Click(object sender, EventArgs e)
        {
            // The GenericDefault StringFormat object is a singleton and changes to its flags will be persisted
            // in the object and will be reflected in other places where this object is used. Let's clone it.
            OnStringFormatUpdate(StringFormat.GenericDefault.Clone() as StringFormat); // Same as default ctr.
        }

        /// <summary>
        ///     Set the UStringFormat object to StringFormat.GenericTypographics and forces an update of the
        ///     StringFormat option boxes in the OptionsForm.
        /// </summary>
        private void BtnStringFormatGenericTypographic_Click(object sender, EventArgs e)
        {
            // The GenericTypographic StringFormat object is a singleton and changes to its flags will be persisted
            // in the object and will be reflected in other places where this object is used. Let's clone it.
            OnStringFormatUpdate(StringFormat.GenericTypographic.Clone() as StringFormat);
        }

        /// <summary>
        /// </summary>
        void DrawingSurface_Load(object sender, EventArgs e)
        {
            UpdateDrawingSurfaceSizeTextBox();
        }

        /// <summary>
        /// </summary>
        private void DrawingSurface_SizeChanged(object sender, EventArgs e)
        {
            UpdateDrawingSurfaceSizeTextBox();
        }

        /// <summary>
        ///     Opens the Font picker dialog and updates the DrawingFont property.
        /// </summary>
        private void LinkLabelFont_LinkClicked(object sender, LinkLabelLinkClickedEventArgs e)
        {
            this.fontDialog1.ShowDialog();
            this.DrawingFont = this.fontDialog1.Font;
        }

        /// <summary>
        ///     Shows MSDN documentation for StringFormat options.
        /// </summary>
        private void LinkLabelTextFormatMSDN_LinkClicked(object sender, LinkLabelLinkClickedEventArgs e)
        {
            ShowMSDNLink(new Uri("http://msdn.microsoft.com/library/default.asp?url=/library/en-us/cpref/html/frlrfsystemdrawingstringformatmemberstopic.asp"));
        }

        /// <summary>
        ///     Shows MSDN documentation for TextFormatFlags options.
        /// </summary>
        private void LinkLabelStingFormatMSDN_LinkClicked(object sender, LinkLabelLinkClickedEventArgs e)
        {
            ShowMSDNLink(new Uri("http://msdn.microsoft.com/library/default.asp?url=/library/en-us/gdi/fontext_4pbs.asp"));
        }

        /// <summary>
        /// </summary>
        private void OptionsForm_Load(object sender, EventArgs e)
        {
            OnDrawingFontChanged();
            DrawingSurface.Show();
        }

        // Other methods

        /// <summary>
        ///     Gets a StringFormat object from the specified TextFormatFlags flags.
        /// </summary>
        private static StringFormat GetStringFormatFromTextFormatFlags(TextFormatFlags flags)
        {
            StringFormat sf;

            /*
            if (flags == TextFormatFlags.Default)
            {
                //  A generic, default StringFormat object has the following characteristics: 
                //   No string format flags are set. 
                //   Character alignment and line alignment are set to StringAlignmentNear. 
                //   Language ID is set to neutral language, which means that the current language associated with the calling thread is used. 
                //   String digit substitution is set to StringDigitSubstituteUser. 
                //   Hot key prefix is set to HotkeyPrefixNone. 
                //   Number of tab stops is set to zero. 
                //   String trimming is set to StringTrimmingCharacter. 

                return sf = StringFormat.GenericDefault;
            }
            */            

            // StringFormatFlags.
            // No GDI counterparts for StringFormatFlags:
            // DirectionVertical, DisplayFormatControl, NoFontFallback (almost like TextFormatFlags.Internal).

            if ((flags & TextFormatFlags.NoPadding) != 0)
            {
                // The only way to get no padding in GDI+ is with a GenericTypograhpics string format.
                // For ref: A generic, typographic StringFormat object has the following characteristics: 
                //   String format flags StringFormatFlagsLineLimit, StringFormatFlagsNoClip, and StringFormatFlagsNoFitBlackBox are set. 
                //   Character alignment and line alignment are set to StringAlignmentNear. 
                //   Language ID is set to neutral language, which means that the current language associated with the calling thread is used. 
                //   String digit substitution is set to StringDigitSubstituteUser. 
                //   Hot key prefix is set to HotkeyPrefixNone. 
                //   Number of tab stops is set to zero. 
                //   String trimming is set to StringTrimmingNone. 
                //  Also, the text margins are set to 0 (this is undocumented).
                sf = StringFormat.GenericTypographic.Clone() as StringFormat; // clone, to avoid persisting changes to the shared GenericTypographic format.

                // Clear the flags that we may not want
                sf.FormatFlags &= ~(StringFormatFlags.LineLimit | StringFormatFlags.NoClip);

                // Leave NoFitBlackBox. StringFormatFlags.FitBlackBox actually means NoFitBlackBox. See VSWhidbey#434198.
                //sf.FormatFlags |= StringFormatFlags.FitBlackBox;
            }
            else
            {
                sf = new StringFormat();
            }

            // GDI always includes trailing spaces.
            sf.FormatFlags |= StringFormatFlags.MeasureTrailingSpaces;

            // WordBreak is the default.  
            // In GDI, text layout is honored if SingleLine or WordBreak not specified, i.e.: single-lined text won't break
            // even if it does not fit the binding box, and multi-line text will break as specified in the paragraph, clipping
            // lines that don't fit the rectangle.
            if ((flags & TextFormatFlags.SingleLine) != 0)
            {
                sf.FormatFlags |= StringFormatFlags.NoWrap;
            }
            else
            // TextBoxControl requires WordBreak.  If TextBoxControl is set end-line-word will break (character break),
            // also if last line is partially clippled (height) it is excluded (fully clipped).
            if ((flags & TextFormatFlags.TextBoxControl) != 0 && (flags & TextFormatFlags.WordBreak) != 0)
            {
                sf.FormatFlags |= StringFormatFlags.LineLimit;
            }

            if ((flags & TextFormatFlags.NoClipping) != 0)
            {
                sf.FormatFlags |= StringFormatFlags.NoClip;
            }

            // StringTrimming.
            // No GDI counterparts for StringTrimming:
            // Word, Character (do not add ellipsis).
            // WordEllipsis takes precedence over the others and draws single-lined text.
            if ((flags & TextFormatFlags.WordEllipsis) != 0)  
            {
                sf.Trimming = StringTrimming.EllipsisWord;
                sf.FormatFlags |= StringFormatFlags.NoWrap;
            }
            else if ((flags & TextFormatFlags.EndEllipsis) != 0)
            {
                sf.Trimming = StringTrimming.EllipsisCharacter;
            }
            else if ((flags & TextFormatFlags.PathEllipsis) != 0)
            {
                sf.Trimming = StringTrimming.EllipsisPath;
            }

            // If RTL is enabled, GDI+ interprets Near as GDI would Right
            // and Far as Left for horizontal alignment.
            bool rtl = (flags & TextFormatFlags.RightToLeft) != 0;

            // Support for RTL languages.
            //
            if (rtl)
            {
                sf.FormatFlags |= StringFormatFlags.DirectionRightToLeft;
            }

            // StringAlignment (Horizontal).
            // 
            if ((flags & TextFormatFlags.HorizontalCenter) != 0)
            {
                sf.Alignment = StringAlignment.Center;
            }
            else if ((flags & TextFormatFlags.Right) != 0)
            {
                sf.Alignment = rtl ? StringAlignment.Near : StringAlignment.Far;
            }
            else
            {
                sf.Alignment = rtl ? StringAlignment.Far : StringAlignment.Near;
            }

            // LineAlignment (Vertical).
            // 
            if ((flags & TextFormatFlags.VerticalCenter) != 0)
            {
                sf.LineAlignment = StringAlignment.Center;
            }
            else if ((flags & TextFormatFlags.Bottom) != 0)
            {
                sf.LineAlignment = StringAlignment.Far;
            }
            else
            {
                sf.LineAlignment = StringAlignment.Near;
            }

            // HotkeyPrefix.
            // No HotkeyPrefix for TextFormatFlags.PrifexOnly (does not render the text but only the hotkey indicator).
            if ((flags & TextFormatFlags.NoPrefix) != 0)
            {
                sf.HotkeyPrefix = HotkeyPrefix.None;
            }
            else if ((flags & TextFormatFlags.HidePrefix) != 0)
            {
                sf.HotkeyPrefix = HotkeyPrefix.Hide;
            }
            else
            {
                sf.HotkeyPrefix = HotkeyPrefix.Show;
            }

            // Others 

            if ((flags & TextFormatFlags.ExpandTabs) != 0)
            {
                // The default character size of a tab stop in GDI is 8 (char average) but GDI+ uses logical units.  
                // This needs to be adjusted before drawing the text.
                // 
                sf.SetTabStops(0f, new float[] { 0f });
            }
            //if ((flags & TextFormatFlags.TabStop) != 0) // Not supported.
            //{
            //    sf.SetTabStops(0, null);
            //}

            // No GDI+ counterparts for the following TextFormatFlags:
            // ExternalLeading: Add external leading portion of the font height when measuring.
            // ModifyString: When used with Ellipsis options, it modifies the string buffer to match the displayed string.
            // NoFullWidthCharacterBreak: Uses SBCS word-line-breaking algorithm for DBCS.
            // WordBreak: GDI+ uses bounding rectangles as the indication for this flag.
            // LeftAndRightPadding (non-GDI).

            return sf;
        }

        /// <summary>
        ///     Gets a TextFormatFlags value from the specified StringFormat object.
        /// </summary>
        private static TextFormatFlags GetTextFormatFlagsFromStringFormat(StringFormat format)
        {
            TextFormatFlags flags = TextFormatFlags.Default;

            if (format == null /*|| format.Equals(StringFormat.GenericDefault)*/)
            {
                //  A generic, default StringFormat object has the following characteristics: 
                //  No string format flags are set. 
                //  Character alignment and line alignment are set to StringAlignmentNear. 
                //  Language ID is set to neutral language, which means that the current language associated with the calling thread is used. 
                //  String digit substitution is set to StringDigitSubstituteUser. 
                //  Hot key prefix is set to HotkeyPrefixNone. 
                //  Number of tab stops is set to zero. 
                //  String trimming is set to StringTrimmingCharacter. 

                return flags;
            }

            // StringFormatFlags.

            /*
            if ((format.FormatFlags & StringFormatFlags.DirectionVertical) != 0 )
            {
                // N/A
            }

            if ((format.FormatFlags & StringFormatFlags.DisplayFormatControl) != 0 )
            {
                // N/A
            }

            if ((format.FormatFlags & StringFormatFlags.MeasureTrailingSpaces) != 0 )
            {
                // N/A
            }

            if ((format.FormatFlags & StringFormatFlags.NoFontFallback) != 0 )
            {
                // N/A
            }
            */

            float dummy;
            if (format.GetTabStops(out dummy) != null)
            {
                flags |= TextFormatFlags.ExpandTabs;
            }

            if ((format.FormatFlags & StringFormatFlags.DirectionRightToLeft) != 0)
            {
                flags |= TextFormatFlags.RightToLeft;
            }

            if ((format.FormatFlags & StringFormatFlags.FitBlackBox) != 0)
            {
                // FitBlackBox actually means NoFitBlackBox.
                flags |= TextFormatFlags.NoPadding;
            }

            if ((format.FormatFlags & StringFormatFlags.NoClip) != 0)
            {
                flags |= TextFormatFlags.NoClipping;
            }

            if ((format.FormatFlags & StringFormatFlags.NoWrap) != 0)
            {
                flags |= TextFormatFlags.SingleLine;
            }
            else
            {
                flags |= TextFormatFlags.WordBreak;
            }

            if ((format.FormatFlags & StringFormatFlags.LineLimit) != 0)
            {
                flags |= TextFormatFlags.TextBoxControl | TextFormatFlags.WordBreak;
            }


            // Horizontal alignment.
            switch (format.Alignment)
            {
                case StringAlignment.Center:
                    flags |= TextFormatFlags.HorizontalCenter;
                    break;
                case StringAlignment.Far:
                    flags |= TextFormatFlags.Right;
                    break;
                case StringAlignment.Near:
                default:
                    flags |= TextFormatFlags.Left;
                    break;
            }


            // Vertical alignment.
            switch (format.LineAlignment)
            {
                case StringAlignment.Center:
                    flags |= TextFormatFlags.VerticalCenter;
                    break;
                case StringAlignment.Far:
                    flags |= TextFormatFlags.Bottom;
                    break;
                case StringAlignment.Near:
                default:
                    flags |= TextFormatFlags.Top;
                    break;
            }

            // StringTrimming.
            switch (format.Trimming)
            {
                case StringTrimming.EllipsisCharacter:
                    flags |= TextFormatFlags.EndEllipsis;
                    break;
                case StringTrimming.EllipsisPath:
                    flags |= TextFormatFlags.PathEllipsis;
                    break;
                case StringTrimming.EllipsisWord:
                    flags |= TextFormatFlags.WordEllipsis;
                    break;
                case StringTrimming.None:
                // Default.
                case StringTrimming.Character:
                case StringTrimming.Word:
                // N/A
                default:
                    break;
            }

            // Hotkey prefix
            switch (format.HotkeyPrefix)
            {
                case HotkeyPrefix.Hide:
                    flags |= TextFormatFlags.HidePrefix;
                    break;
                case HotkeyPrefix.None:
                    flags |= TextFormatFlags.NoPrefix;
                    break;
                case HotkeyPrefix.Show:
                default:
                    break;
            }

            return flags;
        }

        /// <summary>
        ///     Creates C# code for the TextFormatFlags selected.
        /// </summary>
        private string GenerateTextFormatFlagsCode()
        {
            string txtFlags = "";

            if( this.flags != TextFormatFlags.Default )
            {
                if( ( this.flags & TextFormatFlags.Bottom ) != 0 )
                {
                    txtFlags += " | TextFormatFlags.Bottom";
                }
                if( ( this.flags & TextFormatFlags.Right ) != 0 )
                {
                    txtFlags += " | TextFormatFlags.Right";
                }
                if( ( this.flags & TextFormatFlags.RightToLeft ) != 0 )
                {
                    txtFlags += " | TextFormatFlags.RightToLeft";
                }
                if( ( this.flags & TextFormatFlags.HorizontalCenter ) != 0 )
                {
                    txtFlags += " | TextFormatFlags.HorizontalCenter";
                }
                if( ( this.flags & TextFormatFlags.VerticalCenter ) != 0 )
                {
                    txtFlags += " | TextFormatFlags.VerticalCenter";
                }

                if( ( this.flags & TextFormatFlags.PreserveGraphicsTranslateTransform ) != 0 )
                {
                    txtFlags += " | TextFormatFlags.PreserveGraphicsTranslateTransform";
                }
                if( ( this.flags & TextFormatFlags.PreserveGraphicsClipping ) != 0 )
                {
                    txtFlags += " | TextFormatFlags.PreserveGraphicsClipping";
                }

                if( ( this.flags & TextFormatFlags.NoPrefix ) != 0 )
                {
                    txtFlags += " | TextFormatFlags.NoPrefix";
                }
                if( ( this.flags & TextFormatFlags.PrefixOnly ) != 0 )
                {
                    txtFlags += " | TextFormatFlags.PrefixOnly";
                }
                if( ( this.flags & TextFormatFlags.HidePrefix ) != 0 )
                {
                    txtFlags += " | TextFormatFlags.HidePrefix";
                }


                if( ( this.flags & TextFormatFlags.NoPadding ) != 0 )
                {
                    txtFlags += " | TextFormatFlags.NoPadding";
                }

                if( ( this.flags & TextFormatFlags.LeftAndRightPadding ) != 0 )
                {
                    txtFlags += " | TextFormatFlags.LeftAndRightPadding";
                }


                if( ( this.flags & TextFormatFlags.WordBreak ) != 0 )
                {
                    txtFlags += " | TextFormatFlags.WordBreak";
                }
                if( ( this.flags & TextFormatFlags.SingleLine ) != 0 )
                {
                    txtFlags += " | TextFormatFlags.SingleLine";
                }
                if( ( this.flags & TextFormatFlags.TextBoxControl ) != 0 )
                {
                    txtFlags += " | TextFormatFlags.TextBoxControl";
                }
                if( ( this.flags & TextFormatFlags.NoFullWidthCharacterBreak ) != 0 )
                {
                    txtFlags += " | TextFormatFlags.NoFullWidthCharacterBreak";
                }


                if( ( this.flags & TextFormatFlags.NoClipping ) != 0 )
                {
                    txtFlags += " | TextFormatFlags.NoClipping";
                }
                if( ( this.flags & TextFormatFlags.ModifyString ) != 0 )
                {
                    txtFlags += " | TextFormatFlags.ModifyString";
                }
                if( ( this.flags & TextFormatFlags.Internal ) != 0 )
                {
                    txtFlags += " | TextFormatFlags.Internal";
                }
                if( ( this.flags & TextFormatFlags.ExternalLeading ) != 0 )
                {
                    txtFlags += " | TextFormatFlags.ExternalLeading";
                }
                if( ( this.flags & TextFormatFlags.ExpandTabs ) != 0 )
                {
                    txtFlags += " | TextFormatFlags.ExpandTabs";
                }


                if( ( this.flags & TextFormatFlags.EndEllipsis ) != 0 )
                {
                    txtFlags += " | TextFormatFlags.EndEllipsis";
                }
                if( ( this.flags & TextFormatFlags.PathEllipsis ) != 0 )
                {
                    txtFlags += " | TextFormatFlags.PathEllipsis";
                }
                if( ( this.flags & TextFormatFlags.WordEllipsis ) != 0 )
                {
                    txtFlags += " | TextFormatFlags.WordEllipsis";
                }
            }

            if( txtFlags == "" )
            {
                txtFlags = "TextFormatFlags.Default";
            }

            if( txtFlags[1] == '|' )
            {
                txtFlags = txtFlags.Substring( 3, txtFlags.Length - 3 );
            }

            txtFlags = "TextFormatFlags textFlags = " + txtFlags + ";";

            return txtFlags;
        }

        /// <summary>
        ///     Creates C# code for the StringFormat selected.
        /// </summary>
        private string GenerateStringFormatCode()
        {
            string strFormat = null;

            if( this.format != null )
            {
                string formatFlags = "";
                strFormat = "StringFormat strFormat = new StringFormat();";

                if( ( format.FormatFlags & StringFormatFlags.DirectionVertical ) != 0 )
                {
                    formatFlags += " | StringFormatFlags.DirectionVertical";
                }
                if( ( format.FormatFlags & StringFormatFlags.DisplayFormatControl ) != 0 )
                {
                    formatFlags += " | StringFormatFlags.DisplayFormatControl";
                }
                if( ( format.FormatFlags & StringFormatFlags.MeasureTrailingSpaces ) != 0 )
                {
                    formatFlags += " | StringFormatFlags.MeasureTrailingSpaces";
                }
                if( ( format.FormatFlags & StringFormatFlags.NoFontFallback ) != 0 )
                {
                    formatFlags += " | StringFormatFlags.NoFontFallback";
                }
                if( ( format.FormatFlags & StringFormatFlags.DirectionRightToLeft ) != 0 )
                {
                    formatFlags += " | StringFormatFlags.DirectionRightToLeft";
                }
                if( ( format.FormatFlags & StringFormatFlags.FitBlackBox ) != 0 )
                {
                    formatFlags += " | StringFormatFlags.FitBlackBox";
                }
                if( ( format.FormatFlags & StringFormatFlags.NoClip ) != 0 )
                {
                    formatFlags += " | StringFormatFlags.NoClip";
                }
                if( ( format.FormatFlags & StringFormatFlags.NoWrap ) != 0 )
                {
                    formatFlags += " | StringFormatFlags.NoWrap";
                }
                if( ( format.FormatFlags & StringFormatFlags.LineLimit ) != 0 )
                {
                    formatFlags += " | StringFormatFlags.LineLimit";
                }

                if( formatFlags != "" )
                {
                    strFormat += "\r\nstrFormat.FormatFlags = " + formatFlags.Substring( 2, formatFlags.Length - 2 ) + ";";
                }


                // Horizontal alignment.
                string hozAlignment = "";
                switch( format.Alignment )
                {
                    case StringAlignment.Center:
                        hozAlignment = "StringAlignment.Center";
                        break;
                    case StringAlignment.Far:
                        hozAlignment = "StringAlignment.Far";
                        break;
                    case StringAlignment.Near:
                    default:
                        break;
                }

                if( hozAlignment != "" )
                {
                    strFormat += "\r\nstrFormat.Alignment = " + hozAlignment;
                }

                // Vertical alignment.
                string lineAlignment = "";
                switch( format.LineAlignment )
                {
                    case StringAlignment.Center:
                        lineAlignment = "StringAlignment.Center";
                        break;
                    case StringAlignment.Far:
                        lineAlignment = "StringAlignment.Far";
                        break;
                    case StringAlignment.Near:
                    default:
                        break;
                }

                if( lineAlignment != "" )
                {
                    strFormat += "\r\nstrFormat.LineAlignment = " + lineAlignment + ";";
                }

                // StringTrimming.
                string trimming = "";
                switch( format.Trimming )
                {
                    case StringTrimming.EllipsisCharacter:
                        trimming = "StringTrimming.EllipsisCharacter";
                        break;
                    case StringTrimming.EllipsisPath:
                        trimming = "StringTrimming.EllipsisPath";
                        break;
                    case StringTrimming.EllipsisWord:
                        trimming = "StringTrimming.EllipsisWord";
                        break;
                    case StringTrimming.Word:
                        trimming = "StringTrimming.Word";
                        break;
                    case StringTrimming.None:
                        trimming = "StringTrimming.None";
                        break;
                    case StringTrimming.Character:
                    default:
                        break;
                }

                if( trimming != "" )
                {
                    strFormat += "\r\nstrFormat.Trimming = " + trimming + ";";
                }

                // Hotkey prefix
                string hotPrefix = "";
                switch( format.HotkeyPrefix )
                {
                    case HotkeyPrefix.Hide:
                        hotPrefix = "HotkeyPrefix.Hide";
                        break;
                    case HotkeyPrefix.Show:
                        hotPrefix = "HotkeyPrefix.Show";
                        break;
                    case HotkeyPrefix.None:
                    default:
                        break;
                }

                if( hotPrefix != "" )
                {
                    strFormat += "\r\nstrFormat.HotkeyPrefix = " + hotPrefix + ";";
                }
            }

            if( strFormat == null )
            {
                strFormat = "StringFormat strFormat = new StringFormat();\r\n";
            }

            return strFormat;
        }

        /// <summary>
        ///     Resets the AutoSize info text boxes.
        /// </summary>
        private void ClearAutoSizeBoxInfo()
        {
            this.txtBoxDrawTextInfo.Text = this.txtBoxDrawStringInfo.Text = "N/A";
        }

        /// <summary>
        ///     Initializes the drawing form.
        /// </summary>
        private void InitializeDrawingForm()
        {
            this.drawingSurface = new SurfaceForm(this.flags, this.format, this.drawingFont, this.drawingText, this.checkBoxAutoSize.Checked, this.checkBoxUseBindingBox.Checked);
            this.drawingSurface.SizeChanged += new EventHandler(DrawingSurface_SizeChanged);
            this.drawingSurface.Load += new EventHandler(DrawingSurface_Load);
            this.drawingSurface.AutoSizeBoxChanged += new SurfaceForm.AutoSizeBoxChangedCallback(DrawingSurface_AutoSizeBoxChanged);
            this.drawingSurface.SuspendLayout();

            this.drawingSurface.Owner = this;
            this.OptionsChanged += new OptionsChangedCallback(this.drawingSurface.OptionsChangedEventHandler);
            this.drawingSurface.ResumeLayout();
        }

        /// <summary>
        ///     Opens a window with a MSDN doc url.
        /// </summary>
        /// <param name="uri"></param>
        private void ShowMSDNLink(Uri uri)
        {
            if (msdn == null || msdn.IsDisposed)
            {
                msdn = new MsdnForm();
            }

            msdn.AutoSize = true;
            msdn.Text = "MSDN Online Library";
            msdn.browser.Url = uri;
            msdn.Show();
        }

        /// <summary>
        ///     Updates the value of the drawing surface size text box.
        /// </summary>
        private void UpdateDrawingSurfaceSizeTextBox()
        {
            Size size = this.drawingSurface.DrawingSurfaceSize;
            this.txtBoxSurfaceSize.Text = size.Width + ", " + size.Height;
        }

        /// <summary>
        ///     Set the StringFormat boxes to the updated StringFormat object.
        /// </summary>
        private void UpdateStringFormatBoxes()
        {
            string flagName;
            RadioButton rb;

            // StringFormatFlag CheckBox controls in the OptionsForm form.
            List<CheckBox> formatFlagsBoxes = new List<CheckBox>();

            this.suspendEvents = true;

            try
            {
                // Reset StringFormat.FormatFlag check boxes.
                foreach (Control c1 in this.groupBoxFormatFlags.Controls)
                {
                    CheckBox cb = c1 as CheckBox;

                    if (cb != null)
                    {
                        cb.Checked = false;
                        formatFlagsBoxes.Add(cb);
                    }
                }

                // Now check the boxes according to the string format.
                foreach (StringFormatFlags stringFormatFlag in Enum.GetValues(typeof(StringFormatFlags)))
                {
                    if ((stringFormatFlag & this.format.FormatFlags) != 0)
                    {
                        flagName = Enum.GetName(typeof(StringFormatFlags), stringFormatFlag);

                        foreach (CheckBox cb in formatFlagsBoxes)
                        {
                            if (cb.Text == flagName)
                            {
                                cb.Checked = true;
                            }
                        }
                    }
                }

                // Alignment
                flagName = this.format.Alignment.ToString();

                foreach (Control c2 in this.groupBoxAligment.Controls)
                {
                    rb = c2 as RadioButton;

                    if (rb != null && c2.Text == flagName)
                    {
                        rb.Checked = true; // this will reset the other flags if needed.
                        break;
                    }
                }

                // LineAlignment
                flagName = this.format.LineAlignment.ToString();

                foreach (Control c2 in this.groupBoxLineAlignment.Controls)
                {
                    rb = c2 as RadioButton;

                    if (rb != null && c2.Text == flagName)
                    {
                        rb.Checked = true; // this will reset the other flags if needed.
                        break;
                    }
                }

                // Trimming
                flagName = this.format.Trimming.ToString();

                foreach (Control c2 in this.groupBoxTrimming.Controls)
                {
                    rb = c2 as RadioButton;

                    if (rb != null && c2.Text == flagName)
                    {
                        rb.Checked = true; // this will reset the other flags if needed.
                        break;
                    }
                }

                // HotkeyPrefix
                flagName = this.format.HotkeyPrefix.ToString();

                foreach (Control c2 in this.groupBoxHotkey.Controls)
                {
                    rb = c2 as RadioButton;

                    if (rb != null && c2.Text == flagName)
                    {
                        rb.Checked = true; // this will reset the other flags if needed.
                        break;
                    }
                }
            }
            finally
            {
                this.suspendEvents = false;
            }

        }

        /// <summary>
        ///     Updates the TextFormatFlags option boxes according to the current TextFormatFlags value.
        /// </summary>
        private void UpdateTextFormatFlagsBoxes()
        {
            List<CheckBox> flagBoxes = new List<CheckBox>();

            this.suspendEvents = true;

            try
            {
                // Get check boxes and reset them.
                foreach (Control c1 in this.groupBoxTextFormatFlags.Controls) // Get flag groups.
                {
                    GroupBox gb = c1 as GroupBox;

                    if (gb != null)
                    {
                        foreach (Control c2 in gb.Controls)
                        {
                            CheckBox cb = c2 as CheckBox;

                            if (cb != null)
                            {
                                cb.Checked = false;
                                flagBoxes.Add(cb);
                            }
                        }
                    }
                }

                // update check boxes.
                foreach (TextFormatFlags flag in Enum.GetValues(typeof(TextFormatFlags)))
                {
                    if ((flag & this.flags) != 0)
                    {
                        string flagName = Enum.GetName(typeof(TextFormatFlags), flag);

                        foreach (CheckBox cb in flagBoxes)
                        {
                            if (cb.Text == flagName)
                            {
                                cb.Checked = true;
                            }
                        }
                    }
                }
            }
            finally
            {
                this.suspendEvents = false;
            }
        }

        /// <summary>
        /// </summary>
        private void txtBoxSurfaceSize_KeyDown(object sender, KeyEventArgs e)
        {
            if (e.KeyCode == Keys.Enter)
            {
                SetDrawingSurfaceSize();
            }
        }

        /// <summary>
        /// </summary>
        private void TxtBoxSurfaceSize_Leave(object sender, EventArgs e)
        {
            SetDrawingSurfaceSize();
        }

        /// <summary>
        ///     Support for RTL languages.
        /// </summary>
        private void CheckBoxInputRtl_CheckedChanged(object sender, EventArgs e)
        {
            this.textBoxDrawingText.RightToLeft = this.checkBoxInputRtl.Checked ? RightToLeft.Yes : RightToLeft.No;
        }


        /// <summary>
        ///     Copy StringFormat to the clipboard as C# code.
        /// </summary>
        private void BtnStringFormatCopy_Click( object sender, EventArgs e )
        {
            string strFormat = GenerateStringFormatCode();

            try
            {
                Clipboard.SetText( strFormat );
            }
            catch
            {
            }

            MessageBox.Show( strFormat, "StringFormat copied to Clipboard!" );
        }

        /// <summary>
        ///     Copy TextFormatFlags to the clipboard as C# code.
        /// </summary>
        private void BtnTextFormatFlagsCopy_Click( object sender, EventArgs e )
        {
            string txtFlags = GenerateTextFormatFlagsCode();

            try
            {
                Clipboard.SetText(txtFlags);
            }
            catch
            {
            }

            MessageBox.Show( txtFlags, "TextFormatFlags copied to Clipboard!" );
        }

        /// <summary>
        ///     Sets the DrawingSurface window size.
        /// </summary>
        private void SetDrawingSurfaceSize()
        {
            try
            {
                this.drawingSurface.DrawingSurfaceSize = (Size)this.sizeConverter.ConvertFrom(this.txtBoxSurfaceSize.Text);
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.Message);
            }
        }
    }
}