using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Drawing.Text;
using System.Text;
using System.Windows.Forms;

namespace TextFormatFlagsUtil
{
    public partial class SurfaceForm : Form
    {
		TextFormatFlags flags;
		StringFormat format;
		Font drawingFont;
		string drawingText;
		bool autosize;
		bool useBindingBox;

        /// <summary>
        ///     The surface text rendering engine.
        /// </summary>
        public enum SurfaceType
        { 
            TextRenderer,
            Graphics
        }

        /// <summary>
        ///     Definition of delegate and event for AutoSizeBoxChanged event.
        /// </summary>
        /// <param name="surfaceType"></param>
        /// <param name="newArea"></param>
        public delegate void AutoSizeBoxChangedCallback(SurfaceType surfaceType, Size newArea);
        private event AutoSizeBoxChangedCallback autoSizeBoxChangedHandlers;

        /// <summary>
        ///     Constructor.
        /// </summary>
        /// <param name="flags"></param>
        /// <param name="sf"></param>
        /// <param name="drawingFont"></param>
        /// <param name="drawingText"></param>
        /// <param name="autoSize"></param>
        /// <param name="useBindingBox"></param>
        public SurfaceForm(TextFormatFlags flags, StringFormat sf, Font drawingFont, string drawingText, bool autoSize, bool useBindingBox)
        {
            InitializeComponent();
			this.flags = flags;
			this.format = sf;
			this.drawingFont = drawingFont;
			this.drawingText = drawingText;
			this.autosize = autoSize;
			this.useBindingBox = useBindingBox;
        }

        /// <summary>
        ///     Handler for the OptionsForm.OptionsChanged event.  When this event happens the drawing surface form
        ///     is repainted with the new values for the options that changed.
        /// </summary>
        /// <param name="changeType"></param>
        /// <param name="newValue"></param>
		public void OptionsChangedEventHandler(OptionsForm.OptionType changeType, object newValue)
		{ 
			switch(changeType)
			{
				case OptionsForm.OptionType.AutoSize:
					this.autosize = (bool) newValue;
					break;
				case OptionsForm.OptionType.Font:
					this.drawingFont = newValue as Font;
					break;
				case OptionsForm.OptionType.StringFormat:
					this.format = newValue as StringFormat;
					break;
				case OptionsForm.OptionType.Text:
					this.drawingText = newValue.ToString();
					break;
				case OptionsForm.OptionType.TextFormatFlags:
					this.flags = (TextFormatFlags)newValue;
					break;
				case OptionsForm.OptionType.UseBindingBox:
					this.useBindingBox = (bool) newValue;
					break;
				default:
					return;
			}

			Refresh();
		}

        /// <summary>
        ///     AutoSizeBoxChanged event.
        /// </summary>
        public event AutoSizeBoxChangedCallback AutoSizeBoxChanged 
        {
            add 
            {
                autoSizeBoxChangedHandlers += value;
            }
            remove 
            {
               autoSizeBoxChangedHandlers -= value;
            }
        }

        /// <summary>
        ///     Event raised when the AutoSizeBox (the text bounding box) has been re-painted.
        /// </summary>
        /// <param name="surfaceType"></param>
        /// <param name="newSize"></param>
        private void OnAutoSizeBoxChanged(SurfaceType surfaceType, Size newSize)
        {
            if (autoSizeBoxChangedHandlers != null)
            {
                autoSizeBoxChangedHandlers(surfaceType, newSize);
            }
        }

        /// <summary>
        ///     Event handler for the TextRenderer panel Paint event.
        /// </summary>
        private void PanelTextRenderer_Paint(object sender, PaintEventArgs e)
        {
            Rectangle panelRect = this.panelTextRenderer.ClientRectangle;

            // GDI+ excludes borders.
			e.Graphics.DrawRectangle(Pens.Red, panelRect.X, panelRect.Y, panelRect.Width - 1, panelRect.Height - 1);

			if (this.autosize)
			{
				Size textSize = new Size(panelRect.Width, panelRect.Height);
				Size clientSize = this.ClientSize;

				if (useBindingBox)
				{
					textSize = TextRenderer.MeasureText(e.Graphics, drawingText, drawingFont, textSize, this.flags);
				}
				else
				{
					textSize = TextRenderer.MeasureText(e.Graphics, drawingText, drawingFont, new Size(int.MaxValue, drawingFont.Height), this.flags);
				}

				panelRect.Width = textSize.Width;
				panelRect.Height = textSize.Height;

				e.Graphics.DrawRectangle(Pens.Blue, panelRect);

                OnAutoSizeBoxChanged(SurfaceType.TextRenderer, panelRect.Size);
			}
			
            this.statusStripTextRenderer.Items[1].Text = panelRect.ToString();

            // GDI includes borders, we don't want to draw over the borders.

			if (useBindingBox)
			{
				TextRenderer.DrawText(e.Graphics, drawingText, drawingFont, panelRect, SystemColors.WindowText, Color.Transparent, this.flags);
			}
			else
			{
                TextRenderer.DrawText(e.Graphics, drawingText, drawingFont, new Point(panelRect.X, panelRect.Y), SystemColors.WindowText, Color.Transparent, this.flags);
			}
        }

        /// <summary>
        ///     Event handler for the Graphics panel Paint event.
        /// </summary>
        private void PanelGraphics_Paint(object sender, PaintEventArgs e)
        {
			Rectangle panelRect = this.panelGraphics.ClientRectangle;
            
            // GDI+ excludes borders.
            e.Graphics.DrawRectangle( Pens.Red, panelRect.X, panelRect.Y, panelRect.Width - 1, panelRect.Height - 1 );

            // This is an approximation to GDI ExpandTabStops.
            float dummy;
            float[] tabStops = this.format.GetTabStops(out dummy);
            if (tabStops != null && tabStops.Length > 0)
            {
                {
                    SizeF size = e.Graphics.MeasureString("m", drawingFont);
                    this.format.SetTabStops(0, new float[] { 3.0f * size.Width });
                }
            }

			if (this.autosize)
			{
				SizeF textSize = new SizeF(panelRect.Width, panelRect.Height);
				Size clientSize = this.ClientSize;

				if (this.useBindingBox)
				{
					textSize = e.Graphics.MeasureString(drawingText, drawingFont, textSize, this.format);
				}
				else
				{
					textSize = e.Graphics.MeasureString(drawingText, drawingFont, new PointF(0,0), this.format);
				}

				panelRect.Width = (int) Math.Ceiling(textSize.Width);
				panelRect.Height = (int) Math.Ceiling(textSize.Height);

				e.Graphics.DrawRectangle(Pens.Blue, panelRect);

                OnAutoSizeBoxChanged(SurfaceType.Graphics, panelRect.Size);
			}

            this.statusStripGraphics.Items[1].Text = panelRect.ToString();

            if (useBindingBox)
            {
                e.Graphics.DrawString(drawingText, drawingFont, SystemBrushes.WindowText, panelRect, this.format);
            }
            else
            {
                e.Graphics.DrawString(drawingText, drawingFont, SystemBrushes.WindowText, panelRect.X, panelRect.Y, this.format);
            }
        }

        /// <summary>
        /// </summary>
        private void SurfaceForm_Resize(object sender, EventArgs e)
        {
            Refresh();
        }

        /// <summary>
        ///     Current size of the drawing surface form.
        /// </summary>
        public Size DrawingSurfaceSize
        {
            get
            {
                return this.panelGraphics != null ? this.panelGraphics.Size : Size.Empty;
            }

            set
            {
                if (this.panelGraphics.Size != value)
                {
                    int widthChange = value.Width - panelGraphics.Width;
                    int heightChange = 2 * (value.Height - panelGraphics.Height);
                    Size formSize = this.Size;

                    formSize.Width += widthChange;
                    formSize.Height += heightChange;

                    this.Size = formSize;

                    this.Refresh();
                }
            }
        }
    }
}